import contextlib
import enum
import sys
import types
from typing import Tuple, Type, TypeVar, Union

from typing_extensions import (
    Annotated,
    Any,
    Final,
    Literal,
    TypeAlias,
    TypeGuard,
    get_args,
    get_origin,
)

_T = TypeVar("_T")


def is_subclass(obj: Any, cls: Type[_T]) -> TypeGuard[Type[_T]]:
    return isinstance(obj, type) and issubclass(obj, cls)


AnnotatedType = type(Annotated[int, lambda x: x > 0])
Unions = (Union, types.UnionType) if sys.version_info >= (3, 10) else (Union,)


def generic_issubclass(cls: Any, par: Union[type, Any, Tuple[type, ...]]) -> bool:
    """检查 cls 是否是 args 中的一个子类, 支持泛型, Any, Union

    Args:
        cls (type): 要检查的类
        par (Union[type, Any, Tuple[type, ...]]): 要检查的类的父类

    Returns:
        bool: 是否是父类
    """
    if par is Any:
        return True
    with contextlib.suppress(TypeError):
        if isinstance(par, AnnotatedType):
            return generic_issubclass(cls, get_args(par)[0])
        if isinstance(par, type):
            return issubclass(cls, par)
        if get_origin(par) in Unions:
            return any(generic_issubclass(cls, p) for p in get_args(par))
        if isinstance(par, TypeVar):
            if par.__constraints__:
                return any(generic_issubclass(cls, p) for p in par.__constraints__)
            if par.__bound__:
                return generic_issubclass(cls, par.__bound__)
        if isinstance(par, tuple):
            return any(generic_issubclass(cls, p) for p in par)
        if isinstance(origin := get_origin(par), type):
            return issubclass(cls, origin)
    return False


def is_union(obj: Any) -> bool:
    return get_origin(obj) in Unions


class _SentinelEnum(enum.Enum):
    _Sentinel = "Sentinel"


Sentinel: Final = _SentinelEnum._Sentinel

Flag: TypeAlias = Literal[Sentinel]
MaybeFlag: TypeAlias = Union[Flag, _T]



"""Saya 相关的工具"""
from __future__ import annotations

import inspect
from datetime import datetime
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    overload,
)

from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.typing import T_Dispatcher
from graia.saya.builtins.broadcast.schema import ListenerSchema
from graia.saya.factory import BufferModifier, SchemaWrapper, buffer_modifier, factory
from graia.scheduler import Timer
from graia.scheduler.saya import SchedulerSchema
from graia.scheduler.timers import (
    crontabify,
    every_custom_hours,
    every_custom_minutes,
    every_custom_seconds,
)
from graia.scheduler.utilles import TimeObject

T_Callable = TypeVar("T_Callable", bound=Callable)
Wrapper = Callable[[T_Callable], T_Callable]

T = TypeVar("T")


def gen_subclass(cls: type[T]) -> Generator[type[T], Any, Any]:
    yield cls
    for sub in cls.__subclasses__():
        yield from gen_subclass(sub)


@buffer_modifier
def dispatch(*dispatcher: T_Dispatcher) -> BufferModifier:
    """附加参数解析器，最后必须接 `listen` 才能起效

    Args:
        *dispatcher (T_Dispatcher): 参数解析器

    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """

    return lambda buffer: buffer.setdefault("dispatchers", []).extend(dispatcher)


@overload
def decorate(*decorator: Decorator) -> Wrapper:
    """附加多个无头装饰器

    Args:
        *decorator (Decorator): 无头装饰器

    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """
    ...


@overload
def decorate(name: str, decorator: Decorator, /) -> Wrapper:
    """给指定参数名称附加装饰器

    Args:
        name (str): 参数名称
        decorator (Decorator): 装饰器

    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """
    ...


@overload
def decorate(map: Dict[str, Decorator], /) -> Wrapper:
    """给指定参数名称附加装饰器

    Args:
        map (Dict[str, Decorator]): 参数名称与装饰器的映射

    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """
    ...


@buffer_modifier
def decorate(*args) -> BufferModifier:
    """给指定参数名称附加装饰器

    Args:
        name (str | Dict[str, Decorator]): 参数名称或与装饰器的映射
        decorator (Decorator): 装饰器

    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """
    arg: Union[Dict[str, Decorator], List[Decorator]]
    if isinstance(args[0], str):
        name: str = args[0]
        decorator: Decorator = args[1]
        arg = {name: decorator}
    elif isinstance(args[0], dict):
        arg = args[0]
    else:
        arg = list(args)

    def wrapper(buffer: Dict[str, Any]) -> None:
        if isinstance(arg, list):
            buffer.setdefault("decorators", []).extend(arg)
        elif isinstance(arg, dict):
            buffer.setdefault("decorator_map", {}).update(arg)

    return wrapper


@buffer_modifier
def priority(priority: int, *events: Type[Dispatchable]) -> BufferModifier:
    """设置事件优先级

    Args:
        priority (int): 事件优先级
        *events (Type[Dispatchable]): 提供时则会设置这些事件的优先级, 否则设置全局优先级

    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """

    def wrapper(buffer: Dict[str, Any]) -> None:
        if events:
            buffer.setdefault("extra_priorities", {}).update((e, priority) for e in events)
        else:
            buffer["priority"] = priority

    return wrapper


@factory
def listen(*event: Union[Type[Dispatchable], str]) -> SchemaWrapper:
    """在当前 Saya Channel 中监听指定事件

    Args:
        *event (Union[Type[Dispatchable], str]): 事件类型或事件名称

    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """
    EVENTS: Dict[str, Type[Dispatchable]] = {e.__name__: e for e in gen_subclass(Dispatchable)}
    events: List[Type[Dispatchable]] = [e if isinstance(e, type) else EVENTS[e] for e in event]

    def wrapper(func: Callable, buffer: Dict[str, Any]) -> ListenerSchema:
        decorator_map: Dict[str, Decorator] = buffer.pop("decorator_map", {})
        buffer["inline_dispatchers"] = buffer.pop("dispatchers", [])
        if decorator_map:
            sig = inspect.signature(func)
            for param in sig.parameters.values():
                if decorator := decorator_map.get(param.name):
                    setattr(param, "_default", decorator)
            func.__signature__ = sig
        return ListenerSchema(listening_events=events, **buffer)

    return wrapper


@factory
def schedule(timer: Union[Timer, str], cancelable: bool = True) -> SchemaWrapper:
    """在当前 Saya Channel 中设置定时任务

    Args:
        timer (Union[Timer, str]): 定时器或者类似 crontab 的定时模板
        cancelable (bool): 是否能够取消定时任务, 默认为 True
    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """

    return lambda _, buffer: SchedulerSchema(
        timer=crontabify(timer) if isinstance(timer, str) else timer, cancelable=cancelable, **buffer
    )


class _TimerProtocol(Protocol):
    def __call__(self, value: int, /, *, base: Optional[TimeObject] = None) -> Generator[datetime, None, None]:
        ...


_TIMER_MAPPING: Dict[str, _TimerProtocol] = {
    "second": every_custom_seconds,
    "minute": every_custom_minutes,
    "hour": every_custom_hours,
}


@factory
def every(
    value: int = 1,
    mode: Literal["second", "minute", "hour"] = "second",
    start: Optional[TimeObject] = None,
    cancelable: bool = True,
) -> SchemaWrapper:
    """在当前 Saya Channel 中设置基本的定时任务

    Args:
        value (int): 时间间隔, 默认为 1
        mode (Literal["second", "minute", "hour"]): 定时模式, 默认为 ’second‘
        start (Optional[Union[datetime, time, str, float]]): 定时起始时间, 默认为 datetime.now()
        cancelable (bool): 是否能够取消定时任务, 默认为 True
    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """

    return lambda _, buffer: SchedulerSchema(
        timer=_TIMER_MAPPING[mode](value, base=start), cancelable=cancelable, **buffer
    )


@factory
def crontab(pattern: str, start: Optional[TimeObject] = None, cancelable: bool = True) -> SchemaWrapper:
    """在当前 Saya Channel 中设置类似于 crontab 模板的定时任务

    Args:
        pattern (str): 类似 crontab 的定时模板
        start (Optional[Union[datetime, time, str, float]]): 定时起始时间, 默认为 datetime.now()
        cancelable (bool): 是否能够取消定时任务, 默认为 True
    Returns:
        Callable[[T_Callable], T_Callable]: 装饰器
    """

    return lambda _, buffer: SchedulerSchema(timer=crontabify(pattern, start), cancelable=cancelable, **buffer)


on_timer = schedule



# pyright: reportMissingImports=false
from typing import Literal

from graia.broadcast import Broadcast

from . import Commander


def create_commander(module: Literal["graia.ariadne", "ichika"]) -> Commander:
    import creart

    broadcast = creart.it(Broadcast)

    if module == "graia.ariadne":
        from graia.ariadne.context import event_ctx
        from graia.ariadne.event.message import MessageEvent

        return Commander(broadcast, event_ctx, MessageEvent)

    elif module == "ichika":
        from ichika.graia import BROADCAST_EVENT
        from ichika.graia.event import MessageEvent

        return Commander(broadcast, BROADCAST_EVENT, MessageEvent)

    else:
        raise ValueError(f"Unknown module: {module}")



from __future__ import annotations

import abc
import functools
import inspect
import re
from contextvars import ContextVar
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    MutableMapping,
    TypeVar,
    Union,
)
from weakref import WeakKeyDictionary, WeakSet

from graia.amnesia.message import Element, MessageChain, Text
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from graia.broadcast.typing import T_Dispatcher
from graia.broadcast.utilles import dispatcher_mixin_handler
from typing_extensions import Self

from .._typing_util import MaybeFlag, Sentinel

T = TypeVar("T")

L_PAREN = ("{", "[")
R_PAREN = ("}", "]")
ESCAPE = {
    "\\": "\x00",
    "[": "\x01",
    "]": "\x02",
    "{": "\x03",
    "}": "\x04",
    "|": "\x05",
}
R_ESCAPE = {v: k for k, v in ESCAPE.items()}


def escape(string: str) -> str:
    """转义字符串

    Args:
        string (str): 要转义的字符串

    Returns:
        str: 转义后的字符串
    """
    for k, v in ESCAPE.items():
        string = string.replace("\\" + k, v)
    string = string.replace("\\", "")
    return string


def unescape(string: str) -> str:
    """逆转义字符串, 自动去除空白符

    Args:
        string (str): 要逆转义的字符串

    Returns:
        str: 逆转义后的字符串
    """
    for k, v in R_ESCAPE.items():
        string = string.replace(k, v)
    return string.strip()


@dataclass(frozen=True, init=False)
class TextFrag:
    __slots__ = "choice"
    choice: frozenset[str]

    def __init__(self, choice: Iterable[str] | str) -> None:
        object.__setattr__(self, "choice", frozenset((choice,) if isinstance(choice, str) else choice))


@dataclass(frozen=True, init=False)
class ParamFrag:
    __slots__ = ("names",)
    names: frozenset[str]

    def __init__(self, names: Iterable[str]) -> None:
        object.__setattr__(self, "names", frozenset(names))


class AnnotatedParam:
    __slots__ = ("name", "annotation", "default", "wildcard")
    name: str
    annotation: str | None
    default: str | None
    wildcard: bool

    def __init__(
        self,
        name: str,
        wildcard: bool = False,
        annotation: str | None = None,
        default: str | None = None,
    ) -> None:
        self.name = name
        self.annotation = annotation
        self.default = default
        self.wildcard = wildcard

    def __repr__(self) -> str:
        return f"AnnotatedParam({'...' if self.wildcard else ''}{self.name}: {self.annotation} = {self.default})"

    def to_param(self) -> ParamFrag:
        return ParamFrag((self.name,))


U_Token = Union[TextFrag, ParamFrag, AnnotatedParam]

ann_assign = re.compile(r"(?P<name>[^:=]+)(?P<annotation>:[^=]+)?(?P<default>=.+)?")


def parse_param(param_str: str) -> ParamFrag | AnnotatedParam:
    wildcard: bool = param_str.startswith("...")
    if wildcard:
        param_str = param_str[3:]
    match = ann_assign.match(param_str)
    assert match, f"Invalid param: {param_str}"
    names, *extra = match.groups()
    names = [unescape(name).strip() for name in names.split("|")]
    if not wildcard and not any(extra):
        return ParamFrag(names)
    assert len(names) == 1, f"Invalid param: {param_str}"
    return AnnotatedParam(names[0], wildcard, *(unescape(s).strip().lstrip(":=").strip() if s else None for s in extra))


def _pop(char_stk: list[str]) -> str:
    piece = "".join(char_stk)
    char_stk.clear()
    return piece


def tokenize(string: str) -> list[U_Token]:
    """将字符串转义化, 并处理为 TextFrag, ParamFrag 两种 token

    Args:
        string (str): 要处理的字符串

    Returns:
        List[Tuple[CommandToken, List[int, str]]]: 处理后的 Token
    """

    string = escape(string)

    paren: str = ""
    char_stk: list[str] = []
    token: list[U_Token] = []
    pop = functools.partial(_pop, char_stk)

    for index, char in enumerate(string):
        if char in L_PAREN + R_PAREN:
            if char in L_PAREN:
                assert not paren, (
                    f"""Duplicated parenthesis character "{char}" @ {index} !"""
                    """Are you sure you've escaped with "\\"?"""
                )
                paren = char
            elif char in R_PAREN:
                piece = pop()
                assert paren, f"No matching parenthesis: {paren} @ {index}"
                if paren == "[":  # CHOICE
                    token.append(TextFrag(unescape(x) for x in piece.split("|")))
                elif paren == "{":  # PARAM
                    token.append(parse_param(piece))
                paren = ""
        elif char == " " and not paren:
            if char_stk:
                token.append(TextFrag(pop()))
        else:
            char_stk.append(char)

    assert not paren, f"Unclosed parenthesis: {paren}"

    if char_stk:
        token.append(TextFrag(pop()))

    return token


class MatchEntry:
    def __init__(self, tokens: list[U_Token]) -> None:
        self.nodes: list[MaybeFlag[frozenset[str]]] = [
            token.choice if isinstance(token, TextFrag) else Sentinel for token in tokens
        ]
        self.tokens: list[TextFrag | ParamFrag] = [
            token.to_param() if isinstance(token, AnnotatedParam) else token for token in tokens
        ]
        self.params: list[ParamFrag] = [token for token in self.tokens if isinstance(token, ParamFrag)]


T_MatchEntry = TypeVar("T_MatchEntry", bound=MatchEntry)


class MatchNode(Generic[T_MatchEntry]):
    __slots__ = ("next", "entries")
    next: dict[MaybeFlag[str], MatchNode[T_MatchEntry]]
    entries: WeakSet[T_MatchEntry]

    def __init__(self) -> None:
        self.next = {}
        self.entries = WeakSet()

    def copy(self) -> Self:
        new_obj = self.__class__()
        new_obj.next = self.next.copy()
        new_obj.entries = self.entries.copy()
        return new_obj

    def push(self, entry: T_MatchEntry, index: int = 0) -> None:
        if index >= len(entry.nodes):
            self.entries.add(entry)
            return
        current: MaybeFlag[frozenset[str]] = entry.nodes[index]
        if current is Sentinel:
            self.next.setdefault(current, MatchNode()).push(entry, index + 1)
        else:
            target_nodes: list[MatchNode[T_MatchEntry]] = []
            conflicts: dict[MatchNode[T_MatchEntry], set[str]] = {}
            for piece, node in {piece: self.next[piece] for piece in current if piece in self.next}.items():
                conflicts.setdefault(node, set()).add(piece)
            for old_node, conflict_fields in conflicts.items():
                new_node = old_node.copy()
                target_nodes.append(new_node)
                for field in conflict_fields:
                    self.next[field] = new_node
                new_node.push(entry, index + 1)
                current -= conflict_fields
            if current:
                new_node = MatchNode()
                new_node.push(entry, index + 1)
                for field in current:
                    self.next[field] = new_node

    def _inspect(self, fwd=""):
        for k, node in self.next.items():
            node._inspect(f"{fwd}{'<PARAM>' if k is Sentinel else k} ")


class raw(abc.ABC):  # wildcard annotation object
    ...


raw.register(MessageChain)


def convert_empty(obj: T) -> MaybeFlag[T]:
    if obj is inspect.Parameter.empty:
        return Sentinel
    return Sentinel if isinstance(obj, Decorator) else obj


class ContextVarDispatcher(BaseDispatcher):
    """分发常量给指定名称的参数"""

    def __init__(self, data_ctx: ContextVar[dict[str, Any]]) -> None:
        self.data_ctx = data_ctx

    async def catch(self, interface: DispatcherInterface):
        return self.data_ctx.get().get(interface.name)


ChainContent = List[Union[str, Element]]

ChainContentList = List[ChainContent]

split_cache: MutableMapping[MessageChain, ChainContentList] = WeakKeyDictionary()

quote_pairs = {"'": "'", '"': '"', "‘": "’", "“": "”"}


def extract_str(buf: ChainContent) -> str | None:
    if len(buf) == 1 and isinstance(buf[0], str):
        return buf[0]


def split(chain: MessageChain) -> ChainContentList:
    if chain in split_cache:
        return split_cache[chain]
    result: ChainContentList = []
    quote: str = ""
    buffer: ChainContent = []

    for elem in chain.content:
        if elem.__class__.__name__ == "Quote":
            continue
        if not isinstance(elem, Text):
            buffer.append(elem)
            continue
        cache: list[str] = []
        skipping: bool = False
        for char in elem.text:
            if char == "\\" or skipping:
                skipping = not skipping
                continue
            if char in quote_pairs and not quote:
                quote = quote_pairs[char]
                continue
            elif char == quote:
                quote = ""
                continue
            if char == " " and (cache or buffer) and not quote:
                if cache:
                    buffer.append("".join(cache))
                    cache.clear()
                if buffer:
                    result.append(buffer)
                    buffer = []  # buffer is "move"d, so DO NOT clear.
            elif quote or char != " ":
                cache.append(char)
        if cache:
            buffer.append("".join(cache))
    if buffer:
        result.append(buffer)
    split_cache[chain] = result
    return result


def const_fn(value: T) -> Callable[[], T]:
    return lambda: value


def gen_subclass(cls: type[T]) -> Iterable[type[T]]:
    yield cls
    for sub in cls.__subclasses__():
        yield from gen_subclass(sub)


def resolve_dispatchers_mixin(dispatchers: Iterable[T_Dispatcher]) -> list[T_Dispatcher]:
    res = []
    for d in dispatchers:
        res.extend(dispatcher_mixin_handler(d))
    return res


GRAIA_AFFILIATED = ["ichika", "graia", "avilla"]


def graia_affiliated(cls: type) -> bool:
    return any(cls.__module__.startswith(x) for x in GRAIA_AFFILIATED)



"""Utilities for Graia Framework Community."""

import contextlib

from .interrupt import AnnotationWaiter as AnnotationWaiter
from .interrupt import EventWaiter as EventWaiter
from .interrupt import FunctionWaiter as FunctionWaiter

with contextlib.suppress(ImportError):
    from .saya import decorate as decorate
    from .saya import dispatch as dispatch
    from .saya import listen as listen
    from .saya import priority as priority

# TODO: message parsers using Amnesia



from __future__ import annotations

import re

from graia.amnesia.message import Element, MessageChain, Text


def chain(elements: list[Element]) -> MessageChain:
    from graia.amnesia import message

    return message.__message_chain_class__(elements)


def text(string: str) -> Text:
    from graia.amnesia import message

    return message.__text_element_class__(string)


def map_chain(chain: MessageChain) -> tuple[str, dict[str, Element]]:
    elem_mapping: dict[str, Element] = {}
    elem_str_list: list[str] = []
    for i, elem in enumerate(chain.content):
        if not isinstance(elem, Text):
            elem_mapping[str(i)] = elem
            elem_str_list.append(f"\x02{i}_{elem.__class__.__name__}\x03")
        else:
            elem_str_list.append(elem.text)
    return "".join(elem_str_list), elem_mapping


__element_pattern = re.compile("(\x02\\w+\x03)")


def unmap_chain(string: str, mapping: dict[str, Element]) -> MessageChain:
    elements: list[Element] = []
    for x in __element_pattern.split(string):
        if x:
            if x[0] == "\x02" and x[-1] == "\x03":
                index, class_name = x[1:-1].split("_")
                if mapping[index].__class__.__name__ != class_name:
                    raise ValueError("Validation failed: not matching element type!")
                elements.append(mapping[index])
            else:
                elements.append(text(x))
    return chain(elements)



"""Commander: 便捷的指令触发体系"""
from __future__ import annotations

import abc
import asyncio
import contextlib
import copy
import inspect
from contextvars import ContextVar
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    Iterable,
    List,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from typing import _type_repr as type_repr

from graia.amnesia.message import Element, MessageChain, Text
from graia.amnesia.message import __message_chain_class__ as _msg_
from graia.amnesia.message import __text_element_class__ as _text_
from graia.broadcast import Broadcast, Dispatchable, Listener
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.exectarget import ExecTarget
from graia.broadcast.exceptions import PropagationCancelled
from graia.broadcast.typing import T_Dispatcher

# TODO: Decouple from PyDantic
from pydantic import BaseConfig, BaseModel
from pydantic.class_validators import Validator
from pydantic.fields import ModelField
from typing_extensions import Self

from .._typing_util import MaybeFlag, Sentinel
from ._util import (
    AnnotatedParam,
    ChainContent,
    ChainContentList,
    ContextVarDispatcher,
    MatchEntry,
    MatchNode,
    ParamFrag,
    TextFrag,
    const_fn,
    convert_empty,
    extract_str,
    gen_subclass,
    graia_affiliated,
    raw,
    resolve_dispatchers_mixin,
    split,
    tokenize,
)

T_Callable = TypeVar("T_Callable", bound=Callable)


def chain_validator(value: Any, field: ModelField) -> Any:
    """MessageChain 处理函数.

    应用作 pydantic 的 Model validator.
    取决于字段类型标注, 若与消息链, 消息元素无关则会直接把消息链用 as_display 转换为字符串.

    Args:
        value (Any): 验证值
        field (ModelField): 当前的 model 字段
    """
    if not isinstance(value, list):
        return field.get_default() if value is None else value
    if not value:
        return field.get_default()
    if issubclass(field.outer_type_, MessageChain):
        return _msg_(value)
    if isinstance(field.outer_type_, type) and issubclass(field.outer_type_, Element):
        assert len(value) == 1
        v = value[0]
        if issubclass(field.outer_type_, Text):
            assert v.__class__ is str
            return _text_(v)
        assert v.__class__ is field.outer_type_
        return v
    value = _msg_(value)
    return str(value) if field.outer_type_ in (bool, str, int) else value


def wildcard_validator(value: ChainContentList, field: ModelField) -> Any:
    if not isinstance(value, list):
        return value
    if field.outer_type_ is raw:
        return _msg_([_text_(" ")]).join(
            [
                _msg_(
                    [_text_(e) if isinstance(e, str) else e for e in content],
                )
                for content in value
            ]
        )
    altered_field = copy.copy(field)
    altered_field.outer_type_ = field.type_
    return [chain_validator(v, altered_field) for v in value] or field.get_default() or []


class ParamDesc(abc.ABC):
    field: ModelField
    dest: str

    @abc.abstractmethod
    def populate_field(self, validators: Iterable[Callable]) -> None:
        """生成 ParamDesc 上的 ModelField.

        Args:
            validators (Iterable[Callable]): 用作 validator 的 Callable 可迭代对象
        """
        ...

    def validate(self, v: Any) -> Any:
        res, err = self.field.validate(v, {self.field.name: v}, loc=self.dest)
        if err:
            raise ValueError(err)
        return res


class _CommanderModelConfig(BaseConfig):
    copy_on_model_validation: bool = False
    arbitrary_types_allowed: bool = True


def _make_field(
    name: str,
    type: Type,
    default_factory: MaybeFlag[Callable[[], Any]],
    validators: Iterable[Callable] = (),
) -> ModelField:
    new_factory = None if default_factory is Sentinel else default_factory
    return ModelField(
        name=name,
        type_=type,
        model_config=_CommanderModelConfig,
        class_validators={
            f"#commander_validator_{i}#": Validator(v, pre=True, always=True) for i, v in enumerate(validators)
        },
        default_factory=new_factory,
        required=new_factory is None,
    )


class Slot(ParamDesc):
    """Slot"""

    def __init__(
        self,
        target: Union[str, int],
        type: MaybeFlag[Type[Any]] = Sentinel,
        default: MaybeFlag[Any] = Sentinel,
        default_factory: MaybeFlag[Callable[[], Any]] = Sentinel,
    ) -> None:
        """构建 Slot.

        Args:
            target (Union[str, int]): Slot 的目标 token 名
            type (type, optional): Slot 的类型
            default (Any, optional): 默认值
            default_factory (Callable[[], Any], optional): 默认值工厂函数
        """
        self.target = str(target)
        self.type: Union[Literal[Sentinel], Type[Any]] = type
        self.is_optional: bool = False
        self.is_wildcard: bool = False
        if self.type == "raw":
            self.type = raw
        self.default_factory: MaybeFlag[Callable[[], Any]] = (
            const_fn(default) if default is not Sentinel else default_factory
        )

    def populate_field(self, validators: Iterable[Callable]) -> None:
        if self.type is Sentinel:
            self.type = MessageChain if self.default_factory is Sentinel else self.default_factory().__class__
        if self.is_wildcard and self.type is not raw and not TYPE_CHECKING:
            self.type = List[self.type]
        self.is_optional = self.is_wildcard or (self.default_factory is not Sentinel)
        self.field = _make_field(
            self.target,
            self.type,
            self.default_factory,
            validators,
        )

    def merge(self, other: Self) -> Self:
        if self.type is Sentinel and other.type is not Sentinel:
            self.type = other.type
        if self.default_factory is Sentinel and other.default_factory is not Sentinel:
            self.default_factory = other.default_factory
        return self

    def __repr__(self) -> str:
        seg: List[str] = [
            "..." if self.is_wildcard else "",
            self.target,
            ": ",
            type_repr(self.type),
            f" = {self.default_factory()}" if self.default_factory is not Sentinel else "",
            f" -> {self.dest}" if self.dest != self.target else "",
        ]
        return f"Slot({''.join(seg)})"


class Arg(ParamDesc):
    """Argument"""

    headers: FrozenSet[str]

    def __init__(
        self,
        pattern: str,
        type: MaybeFlag[Type[Any]] = Sentinel,
        default: MaybeFlag[Any] = Sentinel,
        default_factory: MaybeFlag[Callable[[], Any]] = Sentinel,
    ) -> None:
        """初始化 Arg.

        Args:
            pattern (str): Arg 的匹配模板, 与 `Commander.command` 使用相同语法.
            type (type, optional): Arg 的类型
            default (Any, optional): 默认值
            default_factory (Callable[[], Any], optional): 默认值工厂函数
        """
        self.type: MaybeFlag[Type[Any]] = Sentinel
        self.tags: List[str] = []
        iter_tokens = iter(tokenize(pattern))
        headers = next(iter_tokens)
        assert isinstance(headers, TextFrag), "Required argument pattern!"
        self.headers = headers.choice
        for token in iter_tokens:
            assert isinstance(token, ParamFrag), "Argument pattern can only be presented at header!"
            assert len(token.names) == 1, "Arg param cannot have alias!"
            self.tags.append(next(iter(token.names)))
        nargs = len(self.tags)
        if nargs == 0:
            self.type = bool
            self.default_factory = const_fn(False)
        elif nargs == 1:
            self.type = MessageChain
        if default is not Sentinel:
            self.default_factory = const_fn(default)
        elif default_factory is not Sentinel:
            self.default_factory = default_factory
        if type is not Sentinel:
            self.type = type

    def populate_field(self, validators: Iterable[Callable]) -> None:
        assert self.dest
        assert self.type is not Sentinel, f"{self} don't have an appropriate type!"
        assert self.default_factory is not Sentinel, f"{self} doesn't have default value!"
        self.field = _make_field(self.dest, self.type, Sentinel, validators)

    def update(self, annotation: MaybeFlag[Any], default: MaybeFlag[Any]) -> None:
        if self.type is Sentinel and annotation is not Sentinel:
            self.type = annotation
        if self.default_factory is Sentinel and default is not Sentinel:
            self.default_factory = const_fn(default)

    def __repr__(self) -> str:
        return f"Arg([{'|'.join(self.headers)}]{''.join(f' {{{tag}}}' for tag in self.tags)})"


class CommandEntry(MatchEntry, ExecTarget):
    """命令信息的存储数据结构, 同时可作为 ExecTarget"""

    def __init__(self, priority: int) -> None:
        self.priority = priority
        self.slot_map: Dict[str, Slot] = {}  # Slot.target -> Slot
        self.arg_map: Dict[str, Arg] = {}
        self.targets: Set[str] = set()
        self.header_map: Dict[str, Arg] = {}
        self.extra: Optional[AnnotatedParam] = None
        self._arg_name_map: Optional[Dict[Arg, str]] = None
        self._slot_targets: Optional[Tuple[FrozenSet[str], ...]] = None
        self.optional: List[Slot] = []
        self.wildcard: Optional[Slot] = None

    @property
    def arg_name_map(self) -> Dict[Arg, str]:
        if not self._arg_name_map:
            self._arg_name_map = {v: k for k, v in self.arg_map.items()}
        return self._arg_name_map

    @property
    def slot_targets(self) -> Tuple[FrozenSet[str], ...]:
        if not self._slot_targets:
            self._slot_targets = tuple(
                frozenset(name for name in param.names if name in self.slot_map) for param in self.params
            )
        return self._slot_targets

    def update_from_func(self) -> None:
        """从 ExecTarget.callable 更新 entry 的信息"""
        for name, parameter in inspect.signature(self.callable).parameters.items():
            annotation = convert_empty(parameter.annotation)
            default = convert_empty(parameter.default)
            if name in self.targets:
                parsed_slot = Slot(name, annotation, default)
                parsed_slot.dest = name  # assuming that name is consistent
                self.slot_map.setdefault(name, parsed_slot).merge(parsed_slot)  # parsed slot < provided slot
            if name in self.arg_map:
                self.arg_map[name].update(annotation, default)
            if default is not Sentinel:
                last_token = self.tokens[-1]
                assert isinstance(last_token, ParamFrag), "Expected ParamFrag, not TextFrag!"
                assert (
                    self.slot_map[name].target in last_token.names
                ), "Not setting wildcard / optional on the last slot!"

    def compile_arg(self, compile_result: Dict[str, Any], arg_data: Dict[str, ChainContentList]) -> None:
        for arg in self.arg_name_map:
            if TYPE_CHECKING:
                assert arg.type is not Sentinel
            value = arg.default_factory()
            if arg.dest in arg_data:  # provided in arg_data
                if (
                    len(arg.tags) > 1
                    or issubclass(arg.type, BaseModel)  # user provided a model, then we have to zip it
                    and not graia_affiliated(arg.type)
                ):
                    value = dict(zip(arg.tags, arg_data[arg.dest]))
                elif len(arg.tags):
                    value = arg_data[arg.dest]
                else:  # probably a bool, flip it
                    value = not value
            compile_result[arg.dest] = arg.validate(value)

    def compile_extra(self, compile_result: Dict[str, Any], extra_list: ChainContentList) -> Any:
        for index, slot in enumerate(self.optional):
            compile_result[slot.dest] = slot.validate(extra_list[index])
        if self.wildcard:
            compile_result[self.wildcard.dest] = self.wildcard.validate(extra_list[len(self.optional) :])

    def compile_param(
        self,
        slot_data: Dict[str, ChainContent],  # Slot.target -> ChainContent
        arg_data: Dict[str, ChainContentList],  # Arg.dest -> List[ChainContent]
        extras: ChainContentList,
    ) -> Dict[str, Any]:
        compile_result: Dict[str, Any] = {
            slot.dest: slot.validate(slot_data[target]) for target, slot in self.slot_map.items()
        }
        self.compile_arg(compile_result, arg_data)
        self.compile_extra(compile_result, extras)
        return compile_result


class ParseData(NamedTuple):
    index: int
    node: MatchNode[CommandEntry]
    params: Tuple[ChainContent, ...]


commander_param_ctx = ContextVar("commander_param_ctx")

param_dispatcher = ContextVarDispatcher(commander_param_ctx)


T_Event = TypeVar("T_Event", bound=Dispatchable)


class Commander(Generic[T_Event]):
    """便利的指令触发体系"""

    def __init__(self, broadcast: Broadcast, event_ctx: ContextVar[T_Event], listen: Type[T_Event] | None = None):
        """
        Args:
            broadcast (Broadcast): 事件系统
            listen (bool): 是否监听消息事件
        """
        self.broadcast = broadcast
        self._slot_validators: List[Callable] = [chain_validator]
        self._wildcard_validators: List[Callable] = [wildcard_validator]
        self._arg_validators: List[Callable] = [chain_validator]
        self.match_root: MatchNode[CommandEntry] = MatchNode()
        self.entries: Set[CommandEntry] = set()
        self.event_ctx: ContextVar[T_Event] = event_ctx

        if listen is not None:
            self.broadcast.listeners.append(
                Listener(
                    self.execute,
                    self.broadcast.getDefaultNamespace(),
                    list(gen_subclass(listen)),
                )
            )

    def __del__(self):
        self.broadcast.listeners = [i for i in self.broadcast.listeners if i.callable != self.execute]

    def add_type_cast(self, *caster: Callable, type: Literal["slot", "wildcard", "arg"] = "slot") -> None:
        """添加类型验证器 (type caster / validator)

        Args:
            *caster (Callable): 验证器
            type (Literal["slot", "wildcard", "arg"], optional): \
                应用验证器的区域, 默认为不是 wildcard 的 Slot.
        """
        assert type in ("slot", "wildcard", "arg")
        validators: List[Callable] = getattr(self, f"_{type}_validators")
        validators.extend(caster)

    @staticmethod
    def parse_command(command: str, entry: CommandEntry, nbsp: dict[str, Any]) -> None:
        """从传入的命令补充 entry 的信息

        Args:
            command (str): 命令
            entry (CommandEntry): 命令的 entry
            nbsp (dict[str, Any]): eval 的命名空间
        """
        tokenize_result: List[Union[TextFrag, ParamFrag, AnnotatedParam]] = tokenize(command)
        have_optional: bool = False
        for token in tokenize_result:
            if isinstance(token, TextFrag):
                assert all(
                    pattern not in entry.header_map for pattern in token.choice
                ), f"{token} conflicts with an Arg object!"

            elif isinstance(token, AnnotatedParam):
                assert token.name not in entry.targets, "Duplicated parameter slot!"
                entry.targets.add(token.name)
                parsed_slot = Slot(
                    token.name,
                    eval(
                        token.annotation or "_sentinel",
                        {"raw": raw, "_sentinel": Sentinel, **nbsp},
                    ),
                    eval(token.default or "_sentinel", {"_sentinel": Sentinel, **nbsp}),
                )
                parsed_slot.dest = token.name  # assuming that param_name is consistent
                slot = entry.slot_map.setdefault(token.name, parsed_slot).merge(
                    parsed_slot
                )  # parsed slot < provided slot
                if token.wildcard:
                    assert token is tokenize_result[-1], "Not setting wildcard on the last slot!"
                    slot.is_wildcard = True
                    entry.wildcard = slot
                    continue
                elif slot.default_factory is not Sentinel:  # Definitely an optional
                    have_optional = True
                    continue
            elif isinstance(token, ParamFrag):
                for name in token.names:
                    assert name not in entry.targets, "Duplicated parameter slot!"
                    entry.targets.add(name)
            assert not have_optional, "Optional Slot is mixed with other type of components!"
        MatchEntry.__init__(entry, tokenize_result)

    def command(
        self,
        command: str,
        settings: Optional[Dict[str, Union[Slot, Arg]]] = None,
        dispatchers: Sequence[T_Dispatcher] = (),
        decorators: Sequence[Decorator] = (),
        priority: int = 16,
        *,
        nbsp: Optional[dict[str, Any]] = None,
    ) -> Callable[[T_Callable], T_Callable]:
        """装饰一个命令处理函数

        Args:
            command (str): 要处理的命令
            settings (Dict[str, Union[Slot, Arg]], optional): 参数设置.
            dispatchers (Sequence[T_Dispatcher], optional): 可选的额外 Dispatcher 序列.
            decorators (Sequence[Decorator], optional): 可选的额外 Decorator 序列.
            nbsp (dict[str, Any], optional): 可选的字符串评估命名空间.
        Raises:
            ValueError: 命令格式错误

        Returns:
            Callable[[T_Callable], T_Callable]: 装饰器
        """

        entry = CommandEntry(priority)
        self.entries.add(entry)  # Add strong ref

        for name, val in (settings or {}).items():
            if isinstance(val, Slot):
                entry.slot_map[val.target] = val
            elif isinstance(val, Arg):
                for header in val.headers:
                    entry.header_map[header] = val
                entry.arg_map[name] = val
            else:
                raise TypeError(f"Unknown setting value: {name} - {val!r}")
            val.dest = name

        def wrapper(func: T_Callable) -> T_Callable:
            Commander.parse_command(command, entry, {**func.__globals__, **(nbsp or {})})
            ExecTarget.__init__(
                entry,
                func,
                [
                    *resolve_dispatchers_mixin(dispatchers),
                ],
                list(decorators),
            )
            entry.update_from_func()

            # compute optional slots dynamically
            for token in entry.tokens:
                if isinstance(token, ParamFrag):
                    for token in token.names:
                        if (slot := entry.slot_map.get(token)) and slot.default_factory is not Sentinel:
                            entry.optional.append(slot)
                            break

            # populate fields
            for slot in entry.slot_map.values():
                slot.populate_field(self._wildcard_validators if slot.is_wildcard else self._slot_validators)
            for arg in entry.arg_map.values():
                arg.populate_field(self._arg_validators)
            for optional_key in [k for k, v in entry.slot_map.items() if v.is_optional]:
                entry.slot_map.pop(optional_key)
            for _ in entry.optional:
                entry.nodes.pop()
            if entry.wildcard:
                entry.nodes.pop()  # the last optional / wildcard token should not be on the MatchGraph
            self.match_root.push(entry)
            return func

        return wrapper

    def parse_rest(
        self,
        index: int,
        frags: ChainContentList,
        params: Tuple[ChainContent, ...],
        entry: CommandEntry,
    ) -> Optional[Tuple[CommandEntry, dict]]:
        # walks down optional, wildcard and Arg
        # extract slot data based on entry
        slot_data: Dict[str, ChainContent] = {
            name: chain for targets, chain in zip(entry.slot_targets, params) for name in targets
        }
        # slam all the rest data inside extra_list
        extras: ChainContentList = []
        arg_data: Dict[str, ChainContentList] = {}
        # index frags
        while index < len(frags):
            frag: ChainContent = frags[index]
            if (str_frag := extract_str(frag)) in entry.header_map:
                arg = entry.header_map[str_frag]
                index += 1
                if arg.dest:
                    if arg.dest in arg_data:  # if the arg is already assigned
                        return
                    arg_data[arg.dest] = frags[index : index + len(arg.tags)]
                index += len(arg.tags)
                if index > len(frags):  # failed
                    return None
                continue
            else:
                extras.append(frags[index])
                index += 1
        if not entry.wildcard and len(extras) > len(entry.optional):
            return None
        if len(extras) < len(entry.optional):
            extras.extend([] for _ in range(len(entry.optional) - len(extras)))
        return entry, entry.compile_param(slot_data, arg_data, extras)

    async def execute(self, chain: MessageChain):
        """触发 Commander.

        Args:
            chain (MessageChain): 触发的消息链
        """

        frags = split(chain)
        pending_exec: Dict[int, List[Tuple[CommandEntry, dict]]] = {}
        pending_next: Deque[ParseData] = Deque([ParseData(0, self.match_root, ())])

        dispatchers: List[T_Dispatcher] = [param_dispatcher]

        if event := self.event_ctx.get(None):
            dispatchers += resolve_dispatchers_mixin([event.Dispatcher])

        def push_pending(index: int, nxt: MatchNode[CommandEntry], params: Tuple[ChainContent, ...]):
            for entry in nxt.entries:
                with contextlib.suppress(ValueError):
                    if res := self.parse_rest(index, frags, params, entry):
                        pending_exec.setdefault(res[0].priority, []).append(res)
            pending_next.append(ParseData(index, nxt, params))

        while pending_next:
            params: Tuple[ChainContent, ...]
            index, node, params = pending_next.popleft()
            if index >= len(frags):
                continue
            frag = frags[index]
            index += 1
            if (str_frag := extract_str(frag)) in node.next:
                if TYPE_CHECKING:
                    assert isinstance(str_frag, str)
                nxt = node.next[str_frag]
                push_pending(index, nxt, params)
            if Sentinel in node.next:
                nxt = node.next[Sentinel]
                push_pending(index, nxt, params + (frag,))

        for _, execution in sorted(pending_exec.items()):
            tasks: List[asyncio.Task] = []
            for entry, param in execution:
                commander_param_ctx.set(param)
                tasks.append(self.broadcast.loop.create_task(self.broadcast.Executor(entry, dispatchers)))
            done, _ = await asyncio.wait(tasks)
            for task in done:
                if task.exception() and isinstance(task.exception(), PropagationCancelled):
                    raise PropagationCancelled



"""基于 format string 的消息链格式化器"""
from __future__ import annotations

import string
from typing import Any, Literal, Union

from graia.amnesia.message import Element, MessageChain

from ._util import chain, text

_global_formatter = string.Formatter()


class Formatter:
    """类似于 string.Formatter 的消息链格式化器"""

    format_string: str

    def __init__(self, format_string: str) -> None:
        self.format_string = format_string
        self._fields: tuple[tuple[str, str | None, str | None, str | None], ...] = tuple(
            _global_formatter.parse(format_string)
        )
        for field in self._fields:
            # validation, forbidding the expansion for format spec
            *_, format_spec, _ = field
            if format_spec:
                sub_spec = tuple(_global_formatter.parse(format_spec))
                if len(sub_spec) > 1 or any(sub_spec[1:]):  # Definitely not right spec
                    raise ValueError("Format specification expansion is disallowed, found ")

    @staticmethod
    def _convert_field(value: Any, conversion: None | str) -> Any:
        # do any conversion on the resulting object
        if conversion is None:
            return value
        elif conversion == "s":
            return str(value)
        elif conversion == "r":
            return repr(value)
        elif conversion == "a":
            return ascii(value)
        raise ValueError(f"Unknown conversion specifier {conversion!s}")

    @staticmethod
    def _transform(obj: object) -> list[Element]:
        if isinstance(obj, MessageChain):
            return obj.content
        elif isinstance(obj, (Element, str)):
            return [text(obj) if isinstance(obj, str) else obj]
        else:
            return [text(str(obj))]

    def format(
        self,
        *args: Union[Element, MessageChain, str, Any],
        **kwargs: Union[Element, MessageChain, str, Any],
    ) -> MessageChain:
        """通过初始化时传入的格式字符串 格式化消息链

        Args:
            *args (Union[Element, MessageChain, str, Any]): 格式化时传入的位置参数
            **kwargs (Union[Element, MessageChain, str, Any]): 格式化时传入的关键字参数

        Returns:
            MessageChain: 格式化后的消息链
        """
        result: list[Element] = []
        auto_arg_index: int | Literal[False] = 0
        used_args: set[str] = set()
        for field in self._fields:
            literal_text, field_name, format_spec, conversion = field
            if literal_text:
                result.append(text(literal_text))
            if field_name is None:
                continue
            # if there's a field, output it
            # this is some markup, find the object and do
            #  the formatting

            # handle arg indexing when empty field_names are given.
            if field_name == "":
                if auto_arg_index is False:
                    raise ValueError("cannot switch from manual field specification to automatic field numbering")
                field_name = str(auto_arg_index)
                auto_arg_index += 1
            elif field_name.isdigit():
                if auto_arg_index:
                    raise ValueError("cannot switch from manual field specification to automatic field numbering")
                # disable auto arg incrementing, if it gets
                # used later on, then an exception will be raised
                auto_arg_index = False

            # given the field_name, find the object it references
            #  and the argument it came from
            obj, arg_used = _global_formatter.get_field(field_name, args, kwargs)
            used_args.add(arg_used)

            # do any conversion on the resulting object
            obj = self._convert_field(obj, conversion)

            # format the object and append to the result
            if format_spec:
                obj = format(obj, format_spec)

            result.extend(self._transform(obj))

        return chain(result).merge()


def format_chain(f_string: str, /, *args: Any, **kwargs: Any) -> MessageChain:
    return Formatter(f_string).format(*args, **kwargs)



"""基础的消息处理器, 包括 DetectPrefix 与 DetectSuffix"""
import abc
import difflib
import fnmatch
import re
import weakref
from collections import defaultdict
from typing import (
    ClassVar,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from graia.amnesia.message import Element, MessageChain, Text
from graia.broadcast.builtin.derive import Derive
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.exceptions import ExecutionStop
from graia.broadcast.interfaces.decorator import DecoratorInterface
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from typing_extensions import get_args

from ._typing_util import generic_issubclass, is_subclass, is_union
from ._util import map_chain, unmap_chain


class ChainDecorator(abc.ABC, Decorator, Derive[MessageChain]):
    pre = True

    @abc.abstractmethod
    async def __call__(self, chain: MessageChain, interface: DispatcherInterface) -> Optional[MessageChain]:
        ...

    async def target(self, interface: DecoratorInterface):
        return await self(
            await interface.dispatcher_interface.lookup_param("message_chain", MessageChain, None),
            interface.dispatcher_interface,
        )


class DetectPrefix(ChainDecorator):
    """前缀检测器"""

    def __init__(self, prefix: Union[str, Iterable[str]]) -> None:
        """初始化前缀检测器.

        Args:
            prefix (Union[str, Iterable[str]]): 要匹配的前缀
        """
        self.prefix: List[str] = [prefix] if isinstance(prefix, str) else list(prefix)

    async def __call__(self, chain: MessageChain, _) -> Optional[MessageChain]:
        for prefix in self.prefix:
            if chain.startswith(prefix):
                return chain.removeprefix(prefix).removeprefix(" ")

        raise ExecutionStop


class DetectSuffix(ChainDecorator):
    """后缀检测器"""

    def __init__(self, suffix: Union[str, Iterable[str]]) -> None:
        """初始化后缀检测器.

        Args:
            suffix (Union[str, Iterable[str]]): 要匹配的后缀
        """
        self.suffix: List[str] = [suffix] if isinstance(suffix, str) else list(suffix)

    async def __call__(self, chain: MessageChain, _) -> Optional[MessageChain]:
        for suffix in self.suffix:
            if chain.endswith(suffix):
                return chain.removesuffix(suffix).removesuffix(" ")
        raise ExecutionStop


class ContainKeyword(ChainDecorator):
    """消息中含有指定关键字"""

    def __init__(self, keyword: str) -> None:
        """初始化

        Args:
            keyword (str): 关键字
        """
        self.keyword: str = keyword

    async def __call__(self, chain: MessageChain, _) -> Optional[MessageChain]:
        if self.keyword not in chain:
            raise ExecutionStop
        return chain


class MatchContent(ChainDecorator):
    """匹配字符串 / 消息链"""

    def __init__(self, content: Union[str, MessageChain]) -> None:
        """初始化

        Args:
            content (Union[str, MessageChain]): 匹配内容
        """
        self.content: Union[str, MessageChain] = content

    async def __call__(self, chain: MessageChain, _) -> Optional[MessageChain]:
        if isinstance(self.content, str) and str(chain) != self.content:
            raise ExecutionStop
        if isinstance(self.content, MessageChain) and chain != self.content:
            raise ExecutionStop
        return chain


class MatchRegex(ChainDecorator, BaseDispatcher):
    """匹配正则表达式"""

    def __init__(self, regex: str, flags: re.RegexFlag = re.RegexFlag(0), full: bool = True) -> None:
        """初始化匹配正则表达式.

        Args:
            regex (str): 正则表达式
            flags (re.RegexFlag): 正则表达式标志
            full (bool): 是否要求完全匹配, 默认为 True.
        """
        self.regex: str = regex
        self.flags: re.RegexFlag = flags
        self.pattern = re.compile(self.regex, self.flags)
        self.match_func = self.pattern.fullmatch if full else self.pattern.match

    async def __call__(self, chain: MessageChain, _) -> Optional[MessageChain]:
        if not self.match_func(str(chain)):
            raise ExecutionStop
        return chain

    async def beforeExecution(self, interface: DispatcherInterface):
        chain: MessageChain = await interface.lookup_param("message_chain", MessageChain, None)
        _mapping_str, _map = map_chain(chain)
        if res := self.match_func(_mapping_str):
            interface.local_storage["__parser_regex_match_obj__"] = res
            interface.local_storage["__parser_regex_match_map__"] = _map
        else:
            raise ExecutionStop

    async def catch(self, interface: DispatcherInterface):
        if interface.annotation is re.Match:
            return interface.local_storage["__parser_regex_match_obj__"]


class RegexGroup(Decorator):
    """正则表达式组的标志
    以 `Annotated[MessageChain, RegexGroup("xxx")]` 的形式使用,
    或者作为 Decorator 使用.
    """

    def __init__(self, target: Union[int, str]) -> None:
        """初始化

        Args:
            target (Union[int, str]): 目标的组名或序号
        """
        self.assign_target = target

    async def __call__(self, _, interface: DispatcherInterface):
        _res: re.Match = interface.local_storage["__parser_regex_match_obj__"]
        match_group: Tuple[str] = _res.groups()
        match_group_dict: Dict[str, str] = _res.groupdict()
        origin: Optional[str] = None
        if isinstance(self.assign_target, str) and self.assign_target in match_group_dict:
            origin = match_group_dict[self.assign_target]
        elif isinstance(self.assign_target, int) and self.assign_target < len(match_group):
            origin = match_group[self.assign_target]

        return (
            unmap_chain(origin, interface.local_storage["__parser_regex_match_map__"]) if origin is not None else None
        )

    async def target(self, interface: DecoratorInterface):
        return self("", interface.dispatcher_interface)


class MatchTemplate(ChainDecorator):
    """模板匹配"""

    def __init__(self, template: List[Union[Type[Element], Element, str]]) -> None:
        """初始化

        Args:
            template (List[Union[Type[Element], Element]]): 匹配模板， 可以为 `Element` 类或其 `Union`, `str`, `Plain` 实例
        """
        self.template: List[Union[Tuple[Type[Element], ...], Element, str]] = []
        for pattern in template:
            if is_subclass(pattern, Text):
                pattern = "*"
            if isinstance(pattern, type):
                self.template.append((pattern,))
            elif is_union(pattern):  # Union
                assert not any(is_subclass(t, Text) for t in get_args(pattern)), "Leaving Text here leads to ambiguity"
                self.template.append(get_args(pattern))
            elif isinstance(pattern, Element) and not isinstance(pattern, Text):
                self.template.append(pattern)
            else:
                pattern = (
                    re.escape(pattern.text)
                    if isinstance(pattern, Text)
                    else fnmatch.translate(pattern)[:-2]  # truncating the ending \Z
                )
                if self.template and isinstance(self.template[-1], str):
                    self.template[-1] += pattern
                else:
                    self.template.append(pattern)

    def match(self, chain: MessageChain):
        """匹配消息链"""
        if len(self.template) != len(chain):
            return False
        for element, template in zip(chain, self.template):
            if isinstance(template, tuple) and not isinstance(element, template):
                return False
            elif isinstance(template, Element) and element != template:
                return False
            elif isinstance(template, str):
                if not isinstance(element, Text) or not re.match(template, element.text):
                    return False
        return True

    async def __call__(self, chain: MessageChain, _) -> Optional[MessageChain]:
        if not self.match(chain):
            raise ExecutionStop
        return chain


class FuzzyMatch(ChainDecorator):
    """模糊匹配

    Warning:
        我们更推荐使用 FuzzyDispatcher 来进行模糊匹配操作, 因为其具有上下文匹配数量限制.
    """

    def __init__(self, template: str, min_rate: float = 0.6) -> None:
        """初始化

        Args:
            template (str): 模板字符串
            min_rate (float): 最小匹配阈值
        """
        self.template: str = template
        self.min_rate: float = min_rate

    def match(self, chain: MessageChain):
        """匹配消息链"""
        text_frags: List[str] = []
        for element in chain:
            if isinstance(element, Text):
                text_frags.append(element.text)
            else:
                text_frags.append(str(element))
        text = "".join(text_frags)
        matcher = difflib.SequenceMatcher(a=text, b=self.template)
        # return false when **any** ratio calc falls undef the rate
        if matcher.real_quick_ratio() < self.min_rate:
            return False
        if matcher.quick_ratio() < self.min_rate:
            return False
        return matcher.ratio() >= self.min_rate

    async def __call__(self, chain: MessageChain, _) -> Optional[MessageChain]:
        if not self.match(chain):
            raise ExecutionStop
        return chain


class FuzzyDispatcher(BaseDispatcher):
    scope_map: ClassVar[DefaultDict[str, List[str]]] = defaultdict(list)
    event_ref: ClassVar["Dict[int, Dict[str, Tuple[str, float]]]"] = {}

    def __init__(self, template: str, min_rate: float = 0.6, scope: str = "") -> None:
        """初始化

        Args:
            template (str): 模板字符串
            min_rate (float): 最小匹配阈值
            scope (str): 作用域
        """
        self.template: str = template
        self.min_rate: float = min_rate
        self.scope: str = scope
        self.scope_map[scope].append(template)

    async def beforeExecution(self, interface: DispatcherInterface):
        event = interface.event
        if id(event) not in self.event_ref:
            chain: MessageChain = await interface.lookup_param("message_chain", MessageChain, None)
            text_frags: List[str] = []
            for element in chain:
                if isinstance(element, Text):
                    text_frags.append(element.text)
                else:
                    text_frags.append(str(element))
            text = "".join(text_frags)
            matcher = difflib.SequenceMatcher()
            matcher.set_seq2(text)
            rate_calc = self.event_ref[id(event)] = {}
            weakref.finalize(event, lambda d: self.event_ref.pop(d), id(event))
            for scope, templates in self.scope_map.items():
                max_match: float = 0.0
                for template in templates:
                    matcher.set_seq1(template)
                    if matcher.real_quick_ratio() < max_match:
                        continue
                    if matcher.quick_ratio() < max_match:
                        continue
                    if matcher.ratio() < max_match:
                        continue
                    rate_calc[scope] = (template, matcher.ratio())
                    max_match = matcher.ratio()
        win_template, win_rate = self.event_ref[id(event)].get(self.scope, (self.template, 0.0))
        if win_template != self.template or win_rate < self.min_rate:
            raise ExecutionStop

    async def catch(self, i: DispatcherInterface) -> Optional[float]:
        event = i.event
        _, rate = self.event_ref[id(event)].get(self.scope, (self.template, 0.0))
        if generic_issubclass(float, i.annotation) and "rate" in i.name:
            return rate


StartsWith = DetectPrefix
EndsWith = DetectSuffix



"""Broadcast Interrupt 相关的工具"""
from __future__ import annotations

import asyncio
import contextlib
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    cast,
    overload,
)

from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.exceptions import ExecutionStop
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from graia.broadcast.interrupt import InterruptControl, Waiter
from graia.broadcast.typing import T_Dispatcher
from typing_extensions import Self, TypeVarTuple, Unpack

T = TypeVar("T")

T_E = TypeVar("T_E", bound=Dispatchable)


class _ExtendedWaiter(Waiter, Generic[T, T_E]):
    """集成 InterruptControl 的 waiter."""

    listening_events: List[Type[T_E]]

    def __init__(
        self,
        events: List[Type[T_E]],
        dispatchers: List[T_Dispatcher],
        decorators: List[Decorator],
        priority: int,
        block_propagation: bool,
    ) -> None:
        self.listening_events = self.events = events
        self.using_dispatchers = self.dispatchers = dispatchers
        self.using_decorators = self.decorators = decorators
        self.priority = priority
        self.block_propagation = block_propagation

    @overload
    async def wait(self, timeout: float, default: T) -> T:
        ...

    @overload
    async def wait(self, timeout: float, default: Optional[T] = None) -> Optional[T]:
        ...

    @overload
    async def wait(self, timeout: None = None) -> T:
        ...

    async def wait(self, timeout: Optional[float] = None, default: Optional[T] = None):
        """等待 Waiter, 如果超时则返回默认值

        Args:
            timeout (float, optional): 超时时间, 单位为秒
            default (T, optional): 默认值
        """
        from creart import it

        inc = it(InterruptControl)
        if timeout:
            try:
                return await inc.wait(self, timeout=timeout)
            except asyncio.TimeoutError:
                return default
        return await inc.wait(self)


class FunctionWaiter(_ExtendedWaiter[T, Dispatchable]):
    """将 Waiter.create_using_function 封装了一层"""

    def __init__(
        self,
        func: Callable[..., Awaitable[Optional[T]]],
        events: List[Type[Dispatchable]],
        dispatchers: Optional[List[T_Dispatcher]] = None,
        decorators: Optional[List[Decorator]] = None,
        priority: int = 15,
        block_propagation: bool = False,
    ) -> None:
        """
        Args:
            func (Callable): 调用函数
            events (List[Type[Dispatchable]]): 事件类型
            dispatchers (Optional[List[T_Dispatcher]]): 广播器
            decorators (Optional[List[Decorator]]): 装饰器
            priority (int): 优先级
            block_propagation (bool): 是否阻止事件往下传播
        """
        super().__init__(events, dispatchers or [], decorators or [], priority, block_propagation)
        self.detected_event = func  # type: ignore


class EventWaiter(_ExtendedWaiter[T_E, T_E]):
    """将 Waiter.create_using_event 封装了一层."""

    def __init__(
        self,
        events: List[Type[T_E]],
        dispatchers: Optional[List[T_Dispatcher]] = None,
        decorators: Optional[List[Decorator]] = None,
        extra_validator: Optional[Callable[[T_E], bool]] = None,
        priority: int = 15,
        block_propagation: bool = False,
    ) -> None:
        """
        Args:
            events (List[Type[T_E]]): 事件类型
            dispatchers (Optional[List[T_Dispatcher]], optional): Dispatcher 列表
            decorators (Optional[List[Decorator]], optional): Decorator 列表
            extra_validator (Optional[Callable[[T_E], bool]], optional): 额外的验证器
            priority (int, optional): 优先级, 越小越靠前
            block_propagation (bool): 是否阻止事件往下传播
        """
        super().__init__(events, dispatchers or [], decorators or [], priority, block_propagation)
        self.extra_validator = extra_validator

    async def detected_event(self, ev: Dispatchable) -> T_E:
        event = cast(T_E, ev)
        if self.extra_validator and not self.extra_validator(event):
            raise ExecutionStop
        return event


Ts = TypeVarTuple("Ts")
V = TypeVar("V")


class AnnotationWaiter(_ExtendedWaiter[T, T_E]):
    """用于直接获取对应标注的 Waiter."""

    def __init__(
        self,
        annotation: Type[T],
        events: List[Type[T_E]],
        dispatchers: Optional[List[T_Dispatcher]] = None,
        decorator: Optional[Decorator] = None,
        headless_decorators: Optional[List[Decorator]] = None,
        priority: int = 15,
        block_propagation: bool = False,
    ) -> None:
        """
        Args:
            annotation (Type[T]): 参数标注
            events (List[Type[T_E]]): 事件类型
            dispatchers (Optional[List[T_Dispatcher]], optional): Dispatcher 列表
            decorator (Decorator, optional): 可选的参数装饰器
            headless_decorators (Optional[List[Decorator]], optional): 无头 Decorator 列表
            priority (int, optional): 优先级, 越小越靠前
            block_propagation (bool): 是否阻止事件往下传播
        """
        super().__init__(events, dispatchers or [], headless_decorators or [], priority, block_propagation)
        self.annotation: Type[T] = annotation
        self.decorator = decorator
        self.anno_validators: list[tuple[type, Callable[[Any], bool]]] = []

    @overload
    def validator(self, func: Callable[[T_E], bool], /) -> Self:
        ...

    @overload
    def validator(self, annotation: Type[V], func: Callable[[V], bool], /) -> Self:
        ...

    def validator(self, *args: Any) -> Self:
        if len(args) == 1 and callable(args[0]):
            self.anno_validators.append((Dispatchable, args[0]))
        elif len(args) == 2:
            self.anno_validators.append(args)
        else:
            raise ValueError(f"Unrecognized arguments: {args}")
        return self

    async def detected_event(self, dii: DispatcherInterface) -> T:
        for anno, validator in self.anno_validators:
            res = False
            with contextlib.suppress(Exception):
                val: Any = await dii.lookup_param("__AnnotationWaiter_annotation__", anno, None)
                res = validator(val)
            if not res:
                raise ExecutionStop

        return await dii.lookup_param("__AnnotationWaiter_annotation__", self.annotation, self.decorator)
